pub type Color = Colour;

#[derive(Clone, Copy, Debug, PartialEq)]
pub struct Colour([u8; 4]);

impl Colour {
    pub const fn rgba(r: u8, g: u8, b: u8, a: u8) -> Self {
        Self([r, g, b, a])
    }

    pub const fn rgb(r: u8, g: u8, b: u8) -> Self {
        Self([r, g, b, 255])
    }

    pub(crate) fn from_le_bytes(b: &[u8]) -> Self {
        Self(b[0..4].try_into().unwrap())
    }

    pub(crate) fn as_le_bytes(&self) -> [u8; 4] {
        self.0
    }

    pub const fn r(&self) -> u8 {
        self.0[0]
    }
    pub const fn with_r(mut self, r: u8) -> Self {
        self.0[0] = r;
        self
    }

    pub const fn g(&self) -> u8 {
        self.0[1]
    }
    pub const fn with_g(mut self, g: u8) -> Self {
        self.0[1] = g;
        self
    }

    pub const fn b(&self) -> u8 {
        self.0[2]
    }
    pub const fn with_b(mut self, b: u8) -> Self {
        self.0[2] = b;
        self
    }

    pub const fn a(&self) -> u8 {
        self.0[3]
    }
    pub const fn with_a(mut self, a: u8) -> Self {
        self.0[3] = a;
        self
    }
}

/// Colour constants
impl Colour {
    pub const BLACK: Colour = Self::rgba(0, 0, 0, 255);
    pub const WHITE: Colour = Self::rgba(255, 255, 255, 255);
    pub const RED: Colour = Self::rgba(255, 0, 0, 255);
    pub const GREEN: Colour = Self::rgba(0, 255, 0, 255);
    pub const BLUE: Colour = Self::rgba(0, 0, 255, 255);
    pub const TRANSPARENT_GREY: Colour = Self::rgba(128, 128, 128, 128);
}

/// Describes a blending mode between a backdrop and a source pixel.
pub enum BlendMode {
    /// Use source colour without doing any blending.
    SourceCopy,
    /// Use the alpha value of both the source and backdrop pixels.
    Simple,
    /// Only use the alpha value of the source.
    SourceAlpha,
    /// Only use the alpha value of the backdrop.
    BackdropAlpha,
    /// Multiply colour channels together.
    Multiply,
}

impl BlendMode {
    // computes a*alpha + b*(1-alpha)
    const fn mix(a: u8, b: u8, alpha: u8) -> u8 {
        (((a as u16 * alpha as u16) + (b as u16 * (255 - alpha as u16))) / 255) as u8
    }

    // computes a*alpha1 + b*alpha2*(1-alpha)
    const fn mix2(a: u8, b: u8, alpha1: u8, alpha2: u8) -> u8 {
        let a_term = a as u32 * alpha1 as u32 * 255;
        let b_term = b as u32 * alpha2 as u32 * (255 - alpha1) as u32;
        ((a_term + b_term) / (255 * 255)) as u8
    }

    pub const fn blend(&self, source: &Colour, backdrop: &Colour) -> Colour {
        // Alpha compositing colour and alpha formulae: [1]
        // - co = Cs x αs + Cb x αb x (1 - αs)
        // - αo = αs + αb x (1 - αs)
        // [1]: https://www.w3.org/TR/compositing/#simplealphacompositing
        match self {
            Self::SourceCopy => *source,
            Self::Simple => Colour::rgba(
                Self::mix2(source.r(), backdrop.r(), source.a(), backdrop.a()),
                Self::mix2(source.g(), backdrop.g(), source.a(), backdrop.a()),
                Self::mix2(source.b(), backdrop.b(), source.a(), backdrop.a()),
                Self::mix(255, backdrop.a(), source.a()),
            ),
            Self::SourceAlpha => Colour::rgba(
                Self::mix(source.r(), backdrop.r(), source.a()),
                Self::mix(source.g(), backdrop.g(), source.a()),
                Self::mix(source.b(), backdrop.b(), source.a()),
                Self::mix(255, backdrop.a(), source.a()),
            ),
            Self::BackdropAlpha => Colour::rgba(
                Self::mix(source.r(), backdrop.r(), backdrop.a()),
                Self::mix(source.g(), backdrop.g(), backdrop.a()),
                Self::mix(source.b(), backdrop.b(), backdrop.a()),
                Self::mix(255, source.a(), backdrop.a()),
            ),
            Self::Multiply => Colour::rgba(
                ((source.r() as u16 * backdrop.r() as u16) / 255) as u8,
                ((source.g() as u16 * backdrop.g() as u16) / 255) as u8,
                ((source.b() as u16 * backdrop.b() as u16) / 255) as u8,
                ((source.a() as u16 * backdrop.a() as u16) / 255) as u8,
            ),
        }
    }
}

#[test]
fn test_simple_blending() {
    // Opaque white over opaque black
    assert_eq!(
        BlendMode::Simple.blend(&Colour::WHITE, &Colour::BLACK),
        Colour::WHITE
    );

    // Opaque black over opaque white
    assert_eq!(
        BlendMode::Simple.blend(&Colour::BLACK, &Colour::WHITE),
        Colour::BLACK
    );

    // Transparent grey over opaque red
    assert_eq!(
        BlendMode::Simple.blend(&Colour::TRANSPARENT_GREY, &Colour::RED),
        Colour::rgba(191, 64, 64, 255)
    );

    // Transparent grey over transparent grey
    assert_eq!(
        BlendMode::Simple.blend(&Colour::TRANSPARENT_GREY, &Colour::TRANSPARENT_GREY),
        Colour::rgba(96, 96, 96, 191)
    );
}

#[test]
fn test_source_blending() {
    // Opaque white over opaque black
    assert_eq!(
        BlendMode::SourceAlpha.blend(&Colour::WHITE, &Colour::BLACK),
        Colour::WHITE
    );

    // Opaque black over opaque white
    assert_eq!(
        BlendMode::SourceAlpha.blend(&Colour::BLACK, &Colour::WHITE),
        Colour::BLACK
    );

    // Transparent grey over opaque red
    assert_eq!(
        BlendMode::SourceAlpha.blend(&Colour::TRANSPARENT_GREY, &Colour::RED),
        Colour::rgba(191, 64, 64, 255)
    );

    // Transparent grey over transparent grey
    assert_eq!(
        BlendMode::SourceAlpha.blend(&Colour::TRANSPARENT_GREY, &Colour::TRANSPARENT_GREY),
        Colour::rgba(128, 128, 128, 191)
    );
}
