mod data;
mod ops;

#[cfg(target_endian = "big")]
std::compile_error!("kahlo only works on little-endian targets!");

pub mod colour;
pub mod formats;

pub mod prelude {
    pub use super::ops::{AlphaPaintable, Readable, RgbaPaintable, Writable};
    pub use super::{BitmapAccess, BitmapMutAccess};
}

pub mod math {
    pub struct PixelSpace;
    pub type PixelPoint = euclid::Point2D<i32, PixelSpace>;
    pub type PixelSize = euclid::Size2D<i32, PixelSpace>;
    pub type PixelRect = euclid::Rect<i32, PixelSpace>;
    pub type PixelBox = euclid::Box2D<i32, PixelSpace>;
    pub type PixelVector = euclid::Vector2D<i32, PixelSpace>;
    pub type PixelSideOffsets = euclid::SideOffsets2D<i32, PixelSpace>;
}

use formats::PixelFormat;

pub trait BitmapAccess<Format: PixelFormat> {
    fn data(&self) -> &[u8];
    fn width(&self) -> usize;
    fn height(&self) -> usize;

    fn size(&self) -> math::PixelSize {
        math::PixelSize::new(self.width() as i32, self.height() as i32)
    }

    fn row_stride(&self) -> usize;
}

pub trait BitmapMutAccess<Format: PixelFormat>: BitmapAccess<Format> {
    fn data_mut(&mut self) -> &mut [u8];
}

/// Owned bitmap with static pixel format information.
#[derive(Clone, PartialEq)]
pub struct Bitmap<Format: PixelFormat> {
    data: data::OwnedBitmapData,
    _ghost: std::marker::PhantomData<Format>,
}

impl<Format: PixelFormat> Bitmap<Format> {
    pub fn new(width: usize, height: usize) -> Self {
        Self {
            data: data::OwnedBitmapData {
                data: vec![0; width * height * Format::PIXEL_WIDTH],
                width,
                height,
                stride: width * Format::PIXEL_WIDTH,
            },
            _ghost: Default::default(),
        }
    }

    pub fn new_from_vec(data: Vec<u8>, width: usize, height: usize, stride: usize) -> Self {
        Self {
            data: data::OwnedBitmapData {
                data,
                width,
                height,
                stride,
            },
            _ghost: Default::default(),
        }
    }
}

impl<Format: PixelFormat> BitmapAccess<Format> for Bitmap<Format> {
    fn data(&self) -> &[u8] {
        &self.data.data
    }
    fn width(&self) -> usize {
        self.data.width
    }
    fn height(&self) -> usize {
        self.data.height
    }
    fn row_stride(&self) -> usize {
        self.data.stride
    }
}

impl<'l, Format: PixelFormat, T: BitmapAccess<Format>> BitmapAccess<Format> for &'l T {
    fn data(&self) -> &[u8] {
        (*self).data()
    }
    fn size(&self) -> math::PixelSize {
        (*self).size()
    }
    fn width(&self) -> usize {
        (*self).width()
    }
    fn height(&self) -> usize {
        (*self).height()
    }
    fn row_stride(&self) -> usize {
        (*self).row_stride()
    }
}

impl<Format: PixelFormat> BitmapMutAccess<Format> for Bitmap<Format> {
    fn data_mut(&mut self) -> &mut [u8] {
        &mut self.data.data
    }
}

impl<Format: PixelFormat> std::fmt::Debug for Bitmap<Format> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Bitmap")
            .field("width", &self.data.width)
            .field("height", &self.data.height)
            .field("stride", &self.data.stride)
            .finish_non_exhaustive()
    }
}

pub type RgbaBitmap = Bitmap<formats::Rgba32>;
pub type Alphamap = Bitmap<formats::A8>;

/// Borrowed bitmap with static pixel format information.
#[derive(Clone)]
pub struct BitmapRef<'l, Format: PixelFormat> {
    data: data::BorrowedBitmapData<'l>,
    _ghost: std::marker::PhantomData<Format>,
}

impl<'l, Format: PixelFormat> BitmapRef<'l, Format> {
    pub fn new(data: &'l [u8], width: usize, height: usize, stride: usize) -> Self {
        Self {
            data: data::BorrowedBitmapData {
                data,
                width,
                height,
                stride,
            },
            _ghost: Default::default(),
        }
    }
}

impl<'l, Format: PixelFormat> BitmapAccess<Format> for BitmapRef<'l, Format> {
    fn data(&self) -> &[u8] {
        &self.data.data
    }
    fn width(&self) -> usize {
        self.data.width
    }
    fn height(&self) -> usize {
        self.data.height
    }
    fn row_stride(&self) -> usize {
        self.data.stride
    }
}

/// Mutably borrowed bitmap with static pixel format information.
pub struct BitmapMut<'l, Format: PixelFormat> {
    data: data::MutableBitmapData<'l>,
    _ghost: std::marker::PhantomData<Format>,
}

impl<'l, Format: PixelFormat> BitmapAccess<Format> for BitmapMut<'l, Format> {
    fn data(&self) -> &[u8] {
        &self.data.data
    }
    fn width(&self) -> usize {
        self.data.width
    }
    fn height(&self) -> usize {
        self.data.height
    }
    fn row_stride(&self) -> usize {
        self.data.stride
    }
}

impl<'l, Format: PixelFormat> BitmapMutAccess<Format> for BitmapMut<'l, Format> {
    fn data_mut(&mut self) -> &mut [u8] {
        &mut self.data.data
    }
}
